/******************************************************************************\
*                                                                              *
* 15/4/2009                                                                    *
* Tree - Source File                                                           *
*                                                                              *
* (c) Robert Hausmair, Christoph Winklhofer                                    *
*                                                                              *
\******************************************************************************/

//REQUIRED HEADERS 

#include <Graph/Tree.h>
#include <Graph/Graph.h>
#include <Graph/TreeLink.h>
#include <Graph/TreeNodeMovedTo.h>
#include <Graph/NodeSort.h>
#include <TreePlus.h>
#include <Scene.h>
#include <View.h>
#include <Layout/TreeLayout.h>
#include <TreeNodeRep.h>
#include <TreeLinkRep.h>
#include <Animation/Tweener.h>
#include <Animation/NodeTween.h>
#include <Animation/NodeTweenNew.h>
#include <Animation/TweenerNew.h>
#include <Animation/TweenerFinish.h>
#include <Animation/TweenerSetRootStep1.h>
#include <Animation/TweenerSetRootStep2.h>
#include <Animation/TweenerSetRootStep3.h>
#include <Animation/CameraTween.h>
#include <algorithm>
#include <Memory/mmgr.h>
#include <sstream>

//NAMESPACE

using namespace std;

//PRIVATE DEFINITIONS TYPEDEFS

class TreeNodeColorReset : public TreeNodeTransform
{
	virtual void operator()(std::pair<int,TreeNode*> elem);
};

void TreeNodeColorReset::operator()(std::pair<int,TreeNode*> elem)
{
	elem.second->GetNodeRep().SetState(TreeNodeRep::eNORMAL);
}

class TreeNodeAnimationGroupReset : public TreeNodeTransform
{
	virtual void operator()(std::pair<int,TreeNode*> elem);
};

void TreeNodeAnimationGroupReset::operator()(std::pair<int,TreeNode*> elem)
{
	elem.second->SetAnimationGroup(ANIMATE_TREE_FIXED);
}

class TreeeNodeRedraw : public TreeNodeTransform
{
	virtual void operator()(std::pair<int,TreeNode*> elem);
};

void TreeeNodeRedraw::operator()(std::pair<int,TreeNode*> elem)
{
	elem.second->GetNodeRep().SetTitle(QString());
}

/****************************************************************************\
* STATIC                                                                     *
\****************************************************************************/
Tree *Tree::s_Tree = NULL;
//MEMBER

//METHODS

/****************************************************************************\
* CD                                                                         *
\****************************************************************************/

Tree::Tree(Graph &graph, TreePlus &app) : 
	m_graph(graph), 
	m_app(app),
	m_maxDepth(0),
	m_actualNode(NULL),
	m_sortBy(0),
	m_orderBy(0),
	m_animationDuration(1000)
{
	m_pRoot = new TreeNodeImpl(graph.GetRootNode(), *this,  NULL);
	m_pRoot->SetAnimationGroup(ANIMATE_TREE_POSITION_FIX);
	m_nodeCatalog.Insert(m_pRoot->GetId(), m_pRoot);
	m_app.GetScene()->addItem(&m_pRoot->GetNodeRep());
	
	m_level[m_pRoot] = 0;

	//Tweener for node animation
	m_fixedTweener = new Tweener();
	m_fixedTweener->SetDuration(m_animationDuration);
	m_moveTweener = new Tweener();
	m_moveTweener->SetDuration(m_animationDuration);
	m_newTweener = new TweenerNew();
	m_newTweener->SetDuration(m_animationDuration);
	m_finishTweener = new TweenerFinish(m_app.GetView(),m_app.GetSettings(),this);
	m_cameraTweener = new Tweener();
	
	//Tweener to set a new root node
	m_rootTweenerStep1 = new TweenerSetRootStep1();
	m_rootTweenerStep1->SetDuration(m_animationDuration);
	m_rootTweenerStep2 = new TweenerSetRootStep2();
	m_rootTweenerStep2->SetDuration(m_animationDuration);
	m_rootTweenerStep3 = new TweenerSetRootStep3();
	m_rootTweenerStep3->SetDuration(m_animationDuration);
	m_rootTweenerStep1->SetNext(m_rootTweenerStep2);
	m_rootTweenerStep2->SetNext(m_rootTweenerStep3);
	

	static_cast<TweenerNew *>(m_newTweener)->SetExpandedNode(m_pRoot);
	static_cast<TweenerFinish *>(m_finishTweener)->SetExpandedNode(m_pRoot);

	m_bAnimateCamera = true;
	s_Tree = this;
}

Tree::~Tree()
{
	RemoveChildren(m_pRoot);
	//m_app.GetScene()->removeItem(&m_pRoot->GetNodeRep());

	delete m_pRoot;
	delete m_fixedTweener;
	delete m_moveTweener;
	delete m_newTweener;
	delete m_finishTweener;
	delete m_rootTweenerStep1;
	delete m_rootTweenerStep2;
	delete m_rootTweenerStep3;
	delete m_cameraTweener;
}

/****************************************************************************\
* PUBLIC CLASS FUNCTIONS                                                     *
\****************************************************************************/

//EXPLORE NODE
void Tree::ExploreNode( TreeNodeImpl *node )
{
	m_actualNode = node;
	Preview *preview = m_app.GetView()->GetPreview();

	//clear preview
	preview->ClearPreview();

	//clear treeNode coloring
	TreeNodeColorReset colorReset;
	m_nodeCatalog.ForEach(&colorReset);
	node->GetNodeRep().SetState(NodeRep::eEXPLORE);
	
	//get all outgoing nodes
	vector<DataNode*> tmpOutNodes;
	m_graph.GetNodesOut(node->GetDataNode(), tmpOutNodes);

	//get all incoming nodes
	vector<DataNode*> tmpInNodes;
	m_graph.GetNodesIn(node->GetDataNode(), tmpInNodes);

	//set bidirectional nodes
	std::vector<DataNode *> biNodes;
	set_intersection(tmpInNodes.begin(), tmpInNodes.end(), tmpOutNodes.begin(), tmpOutNodes.end(), inserter(biNodes, biNodes.begin()));
	
	//set outgoing nodes (outgoing without bidirectional)
	std::vector<DataNode *> outNodes;
	set_difference(tmpOutNodes.begin(), tmpOutNodes.end(), biNodes.begin(), biNodes.end(), inserter(outNodes, outNodes.begin()));
	
	//set incoming nodes (incoming without bidirectional)
	std::vector<DataNode *> inNodes;
	set_difference(tmpInNodes.begin(), tmpInNodes.end(), biNodes.begin(), biNodes.end(), inserter(inNodes, inNodes.begin()));
	
	//sort nodes
	SortDataNode(outNodes);
	SortDataNode(biNodes);
	SortDataNode(inNodes);
	
	//add outnodes to preview
	for(vector<DataNode*>::iterator it = outNodes.begin(); it != outNodes.end(); it++) 
	{
		//check if already in the tree
		TreeNode *child = m_nodeCatalog.Find((*it)->GetId());
		if(child == NULL) 
		{
			PreviewNodeRep &rep = preview->CreatePreviewNode(false);
			rep.SetTitle((*it)->GetName());
			rep.SetState(NodeRep::eOUTGOING);
			rep.SetId((*it)->GetId());
			SetTooltip(*(*it), rep);
		}
		else 
		{
			PreviewNodeRep &rep = preview->CreatePreviewNode(true);
			rep.SetTitle((*it)->GetName());
			rep.SetState(NodeRep::eOUTGOING);
			rep.SetId((*it)->GetId());
			child->GetNodeRep().SetState(NodeRep::eOUTGOING);
			SetTooltip(*(*it), rep);
		}
	}

	//add inNodes to preview
	for(vector<DataNode*>::iterator it = inNodes.begin(); it != inNodes.end(); it++) 
	{
		//check if already in the tree
		TreeNode *child = m_nodeCatalog.Find((*it)->GetId());
		if(child == NULL) 
		{
			PreviewNodeRep &rep = preview->CreatePreviewNode(false);
			rep.SetTitle((*it)->GetName());
			rep.SetState(NodeRep::eINGOING);
			rep.SetId((*it)->GetId());
			SetTooltip(*(*it), rep);
		}
		else 
		{
			PreviewNodeRep &rep = preview->CreatePreviewNode(true);
			rep.SetTitle((*it)->GetName());
			rep.SetState(NodeRep::eINGOING);
			rep.SetId((*it)->GetId());
			child->GetNodeRep().SetState(NodeRep::eINGOING);
			SetTooltip(*(*it), rep);
		}
	}
	
	//add bidirectional nodes to preview
	for(vector<DataNode*>::iterator it = biNodes.begin(); it != biNodes.end(); it++) 
	{
		//check if already in the tree
		TreeNode *child = m_nodeCatalog.Find((*it)->GetId());
		if(child == NULL) 
		{
			PreviewNodeRep &rep = preview->CreatePreviewNode(false);
			rep.SetTitle((*it)->GetName());
			rep.SetState(NodeRep::eBIDIRECTIONAL);
			rep.SetId((*it)->GetId());
			SetTooltip(*(*it), rep);
		}
		else 
		{
			PreviewNodeRep &rep = preview->CreatePreviewNode(true);
			rep.SetTitle((*it)->GetName());
			rep.SetState(NodeRep::eBIDIRECTIONAL);
			rep.SetId((*it)->GetId());
			child->GetNodeRep().SetState(NodeRep::eBIDIRECTIONAL);
			SetTooltip(*(*it), rep);
		}
	}
}

//EXPAND NODE
void Tree::ExpandNode(TreeNodeImpl* node, bool isResetAnim)
{
	static_cast<TweenerNew *>(m_newTweener)->SetExpandedNode(node);
	static_cast<TweenerFinish *>(m_finishTweener)->SetExpandedNode(node);

	//get outnodes
	vector<DataNode*> outNodes;
	m_graph.GetNodesOut(node->GetDataNode(), outNodes);

	//get innodes
	vector<DataNode*> inNodes;
	m_graph.GetNodesIn(node->GetDataNode(), inNodes);

	//combine in and out nodes in a union
	std::vector<DataNode *> unionNodes;
	set_union(inNodes.begin(), inNodes.end(), outNodes.begin(), outNodes.end(), inserter(unionNodes, unionNodes.begin()));
	
	//get children of this node
	list<TreeNode *> children;
	node->GetChildren(children);

	//count the number the adjacent nodes, without actual node
	int treeNodeChildSize = 0;
	for(vector<DataNode *>::iterator it=unionNodes.begin(); it!=unionNodes.end(); it++) {
		if((*it)->GetId() != node->GetId()) {
			treeNodeChildSize++;
		}
	}

	//get parent nodes
	TreeNodeImpl *parent = static_cast<TreeNodeImpl *>(node->GetParentNode());
	while(parent != NULL)
	{
		if(find(unionNodes.begin(), unionNodes.end(), &parent->GetDataNode()) != unionNodes.end())
		{
			treeNodeChildSize--;
		}
		parent = static_cast<TreeNodeImpl *>(parent->GetParentNode());
	}

	//clear treeNodeAnimationGroups 
	if(isResetAnim) 
	{
		TreeNodeAnimationGroupReset animationReset;
		m_nodeCatalog.ForEach(&animationReset);
	}
	
	int childrenSize = children.size();
	for(list<TreeNode *>::iterator it = children.begin(); it!=children.end(); it++)
	{
		if((*it)->GetType() == TYPE_NODE_MOVEDTO)
		{
			childrenSize--;
			break;
		}
	}

	if(childrenSize == treeNodeChildSize)
	{
		//contract, remove children
		RemoveFromMovedToNode(node);
		RemoveMovedTo(static_cast<TreeNodeImpl *>(m_pRoot));
		
		RemoveChildren(node);
	}
	else
	{
		//expand
		for(vector<DataNode*>::iterator it = unionNodes.begin(); it != unionNodes.end(); it++) 
		{
			TreeNode *child = m_nodeCatalog.Find((*it)->GetId());
			if(child == NULL) 
			{
				child = new TreeNodeImpl(*(*it), *this, node);
				SetLevel(child, m_level[node] + 1);
				node->AddChild(child);

				m_nodeCatalog.Insert(child->GetId(), child);
				m_app.GetScene()->addItem(&child->GetNodeRep());
				m_app.GetScene()->addItem(&child->GetLinkToParent()->GetTreeLinkRep());
				
				//set animationGroup
				child->SetAnimationGroup(ANIMATE_TREE_NEW);
			}
			else if(!IsChildParentOfNode(node, child))
			{
				TreeNodeImpl *oldParent = static_cast<TreeNodeImpl *>(child->GetParentNode());
				child->SetParentNode(node);
				node->AddChild(child);
				SetLevel(child, m_level[node] + 1);
				//set animationGroup
				child->SetAnimationGroup(ANIMATE_TREE_MOVE);

				SetMovedTo(oldParent, static_cast<TreeNodeImpl *>(child));
			}
		}
	}
}

//SET LAYOUT
void Tree::SetLayout()
{
	Sort(m_pRoot);
	
	//reset bound
	m_boundTL = Vector2i(INT_MAX,INT_MAX);
	m_boundBR = Vector2i(INT_MIN,INT_MIN);

	//calculate layout
	TreeLayout layout;
	layout.Position(*this);
}

//SET POSITION
void Tree::SetPosition(TreeNode *node, float x, float y)
{
	//update bound
	if(x < m_boundTL.m_X) m_boundTL.m_X = x;
	if(x > m_boundBR.m_X) m_boundBR.m_X = x;
	if(y < m_boundTL.m_Y) m_boundTL.m_Y = y;
	if(y > m_boundBR.m_Y) m_boundBR.m_Y = y;

	switch(node->GetAnimationGroup())
	{
		case ANIMATE_TREE_FIXED:
		{
			//create Tween
			m_fixedTweener->AddTweenable(new NodeTween(node,node->GetPosition(),Vector2f(x,y)));
			break;
		}
		case ANIMATE_TREE_MOVE:
		{
			m_moveTweener->AddTweenable(new NodeTween(node,node->GetPosition(),Vector2f(x,y)));
			node->GetLinkToParent()->GetTreeLinkRep().hide();
			break;
		}
		case ANIMATE_TREE_NEW:
		{
			//get preview
			Preview *preview = m_app.GetView()->GetPreview();
			
			//get position of preview node 
			PreviewNodeRep *previewRep = preview->GetPreviewNode(node->GetId());

			if(previewRep)
			{
				 m_newTweener->AddTweenable(new NodeTweenNew(node,previewRep,Vector2f(x,y)));
				 node->GetLinkToParent()->GetTreeLinkRep().hide();
			}
			break;
		}
		case ANIMATE_TREE_POSITION_FIX:
		{
			node->SetPosition(x, y);
			break;
		}
	}
}

//UPDATE LAYOUT
void Tree::UpdateLayout()
{
	TreeNodeAnimationGroupReset animationReset;
	m_nodeCatalog.ForEach(&animationReset);
	
	if(GetActualNode() != NULL)
	{
		GetActualNode()->OnNodeClick(&GetActualNode()->GetNodeRep());
		SetLayout();
		StartAnimation(m_pRoot);
		//ExploreNode(GetActualNode());
	}
}

//REDRAW NODES
void Tree::RedrawNodes()
{
	TreeeNodeRedraw redrawT;
	m_nodeCatalog.ForEach(&redrawT);
}


void Tree::SetRootNode(TreeNode *root)
{
	//set new root node
	m_pRoot = root;
	m_pRoot->SetParentNode(NULL);
	this->m_level[m_pRoot] = 0;
	
	map<int, TreeNode*> catalog = m_nodeCatalog.GetAll();

	//move all nodes over the new root node
	float x = root->GetPosition()[0];
	float y = root->GetPosition()[1];
	for(map<int, TreeNode*>::iterator it = catalog.begin(); it != catalog.end(); it++)
	{
		if(it->first != root->GetId())
		{
			m_rootTweenerStep1->AddTweenable(new NodeTween(it->second, it->second->GetPosition(), Vector2f(x,y)));
		}
		if(it->second->GetLinkToParent() != NULL) 
		{
			it->second->GetLinkToParent()->GetTreeLinkRep().hide();
		}
	}
	static_cast<TweenerSetRootStep1 *>(m_rootTweenerStep1)->SetTree(this);
	static_cast<TweenerSetRootStep2 *>(m_rootTweenerStep2)->SetTree(this);
	static_cast<TweenerSetRootStep3 *>(m_rootTweenerStep3)->SetTree(this);

	m_app.GetView()->LockInput(true);
	m_rootTweenerStep3->SetNext(m_finishTweener);
	m_rootTweenerStep1->StartTween();
}

//START ANIMATION
void Tree::StartAnimation(TreeNode *node)
{
	m_app.GetView()->LockInput(true);
	m_app.GetSettings()->setEnabled(false);

	//twenner data
	Tweener *startTweener = NULL;
	int useFixed = 0, useMove = 0, useNew = 0;
	if(m_fixedTweener->GetNmbTweenable())useFixed = 1;
	if(m_moveTweener->GetNmbTweenable())useMove = 2;
	if(m_newTweener->GetNmbTweenable())useNew = 4;

	m_fixedTweener->SetNext(NULL);
	m_moveTweener->SetNext(NULL);
	m_newTweener->SetNext(NULL);

	//cascade tweener
	switch(useFixed | useMove | useNew)
	{
		case 0:
		break;
		
		case 1:
		m_fixedTweener->SetNext(m_finishTweener);
		startTweener = m_fixedTweener;
		break;
		
		case 2:
		m_moveTweener->SetNext(m_finishTweener);
		startTweener = m_moveTweener;
		break;
		
		case 3:
		m_moveTweener->SetNext(m_finishTweener);
		m_fixedTweener->SetNext(m_moveTweener);
		startTweener = m_fixedTweener;
		break;
		
		case 4:
		m_newTweener->SetNext(m_finishTweener);
		startTweener = m_newTweener;
		break;
		
		case 5:
		m_newTweener->SetNext(m_finishTweener);
		m_fixedTweener->SetNext(m_newTweener);
		startTweener = m_fixedTweener;
		break;
		
		case 6:
		m_newTweener->SetNext(m_finishTweener);
		m_moveTweener->SetNext(m_newTweener);
		startTweener = m_moveTweener;
		break;
		
		case 7:
		m_newTweener->SetNext(m_finishTweener);
		m_moveTweener->SetNext(m_newTweener);
		m_fixedTweener->SetNext(m_moveTweener);
		startTweener = m_fixedTweener;
		break;
	}
	
	//set duration
	m_fixedTweener->SetDuration(m_animationDuration);
	m_moveTweener->SetDuration(m_animationDuration);
	m_newTweener->SetDuration(m_animationDuration);
	m_cameraTweener->SetDuration(m_animationDuration);
	m_rootTweenerStep1->SetDuration(m_animationDuration);
	m_rootTweenerStep2->SetDuration(m_animationDuration);
	m_rootTweenerStep3->SetDuration(m_animationDuration);
	
	//start
	if(startTweener)
	{
		//animate camera
		if(m_bAnimateCamera)
		{
			QRectF nodeBound
			( 
				m_boundTL.m_X,
				m_boundTL.m_Y,
				m_boundBR.m_X - m_boundTL.m_X + node->GetNodeRep().Width(), 
				m_boundBR.m_Y - m_boundTL.m_Y + node->GetNodeRep().Height()
			);
			
			QPointF pos = QPointF(node->GetPosition().m_X,node->GetPosition().m_Y);
			m_cameraTweener->AddTweenable(new CameraTween(m_app.GetView(),nodeBound,pos));
			m_cameraTweener->SetNext(startTweener);
			m_cameraTweener->StartTween();
		}
		else
		{
			startTweener->StartTween();
		}
	}
}

bool Tree::IsLeafNode(TreeNode *node)
{
	list<TreeNode *> children;
	node->GetChildren(children);
		
	if(children.size() > 0)
	{
		return false;
	}
	else 
	{
		return true;
	}
}

bool Tree::IsHasLeftSibling(TreeNode *node)
{
	TreeNode *parent = node->GetParentNode();

	// root has no left sibling
	if(parent == NULL)
	{
		return false;
	}
	list<TreeNode *> children;
	parent->GetChildren(children);

	// first node in the list is the left most node.
	// check if the node is the leftmost node.
	if(children.front() != node)
	{
		return true;
	}
	else 
	{
		return false;
	}
}

bool Tree::IsHasRightSibling(TreeNode *node)
{
	TreeNode *parent = node->GetParentNode();

	// root has no left sibling
	if(parent == NULL)
	{
		return false;
	}
	list<TreeNode *> children;
	parent->GetChildren(children);

	// last node in the list is the right most node.
	// check if the node is the rightmost node.
	if(children.back() != node)
	{
		return true;
	}
	else 
	{
		return false;
	}
}

TreeNode * Tree::GetParent(TreeNode *node)
{
	return node->GetParentNode();
}


TreeNode * Tree::GetLeftSibling(TreeNode *node)
{
	TreeNode *parent = node->GetParentNode();
	list<TreeNode *> children;
	parent->GetChildren(children);
	TreeNode *leftSibling = NULL;
	for(list<TreeNode *>::iterator it = children.begin(); it != children.end(); it++)
	{
		if(*it == node)
		{
			break;
		}
		leftSibling = *it;
	}
	return leftSibling;
}

TreeNode * Tree::GetRightSibling(TreeNode *node)
{
	TreeNode *parent = node->GetParentNode();
	list<TreeNode *> children;
	parent->GetChildren(children);
	TreeNode *rightSibling = NULL;
	for(list<TreeNode *>::reverse_iterator it = children.rbegin(); it != children.rend(); it++)
	{
		if(*it == node)
		{
			break;
		}
		rightSibling = *it;
	}
	return rightSibling;
}


void Tree::GetChildren(TreeNode *node, list<TreeNode *>& children)
{
	node->GetChildren(children);
}

TreeNode * Tree::GetFirstChild(TreeNode *node)
{
	list<TreeNode *> children;
	GetChildren(node, children);
	if(children.size() > 0)
	{
		return children.front();
	}
	else 
	{
		return NULL;
	}
}

TreeNode * Tree::GetLastChild(TreeNode *node)
{
	list<TreeNode *> children;
	GetChildren(node, children);
	if(children.size() > 0)
	{
		return children.back();
	}
	else 
	{
		return NULL;
	}
}
	
TreeNode * Tree::GetLeftNeighbor(TreeNode *node)
{
	list<TreeNode *> nodesOnSameLevel;
	GetAllNodesForLevel(this->GetRoot(), m_level[node], nodesOnSameLevel);

	TreeNode *leftNeighbor = NULL;
	for(list<TreeNode *>::iterator it = nodesOnSameLevel.begin(); it != nodesOnSameLevel.end(); it++)
	{
		if(*it == node)
		{
			break;
		}
		leftNeighbor = *it;
	}
	return leftNeighbor;
}

TreeNode * Tree::GetLeftmost(TreeNode *node, int depth)
{
	if(depth == 0) 
	{
		return node;
	}
	else if(IsLeafNode(node))
	{
		return NULL;
	}
	else
	{
		list<TreeNode *> children;
		GetChildren(node, children);
		for(list<TreeNode *>::iterator it=children.begin(); it!= children.end(); it++)
		{
			TreeNode* leftmost = GetLeftmost(*it, depth - 1);
			if(leftmost != NULL)
			{
				return leftmost;
			}
		}
		return NULL;
	}
}

void Tree::GetAllNodesForLevel(TreeNode *node, int level, list<TreeNode *>& nodes)
{
	int nodeLevel = m_level[node];
	if(nodeLevel == level)
	{
		nodes.push_back(node);
	}
	else if(nodeLevel < level)
	{
		list<TreeNode *> children;
		GetChildren(node, children);
		for(list<TreeNode *>::iterator it = children.begin(); it != children.end(); it++)
		{
			GetAllNodesForLevel(*it, level, nodes);
		}
	}
}

void Tree::Sort(TreeNode *node)
{
	if(node->GetType() == TYPE_NODE)
	{
		TreeNodeImpl *nodeImpl = static_cast<TreeNodeImpl *>(node);
		SetTooltip(nodeImpl->GetDataNode(), nodeImpl->GetNodeRep());
	}
	node->SortChildren();
	list<TreeNode *> children;
	node->GetChildren(children);
	for(list<TreeNode *>::iterator it=children.begin(); it != children.end(); it++)
	{
		Sort(*it);
	}
}

void Tree::SortChildren(std::list<TreeNode *>& children)
{
	if(m_orderBy == 0 && m_sortBy == 0)
	{
		children.sort(SortNodeNameAsc);
	}
	else if(m_orderBy == 0 && m_sortBy == 1)
	{
		children.sort(SortNodeNameDesc);
	}
	if(m_orderBy == 1 && m_sortBy == 0)
	{
		children.sort(SortLinkAsc);
	}
	else if(m_orderBy == 1 && m_sortBy == 1)
	{
		children.sort(SortLinkDesc);
	}
	else if(m_orderBy == 2 && m_sortBy == 0)
	{
		children.sort(SortLinkOutAsc);
	}
	else if(m_orderBy == 2 && m_sortBy == 1)
	{
		children.sort(SortLinkOutDesc);
	}
	else if(m_orderBy == 3 && m_sortBy == 0)
	{
		children.sort(SortLinkInAsc);
	}
	else if(m_orderBy == 3 && m_sortBy == 1)
	{
		children.sort(SortLinkInDesc);
	}
	else if(m_orderBy == 4 && m_sortBy == 0)
	{
		children.sort(SortLinkBiAsc);
	}
	else if(m_orderBy == 4 && m_sortBy == 1)
	{
		children.sort(SortLinkBiDesc);
	}
}

void Tree::SortDataNode(std::vector<DataNode *>& nodes)
{
	if(m_orderBy == 0 && m_sortBy == 0)
	{
		sort(nodes.begin(), nodes.end(), SortDataNodeNameAsc);
	}
	else if(m_orderBy == 0 && m_sortBy == 1)
	{
		sort(nodes.begin(), nodes.end(), SortDataNodeNameDesc);
	}
	else if(m_orderBy == 2 && m_sortBy == 0)
	{
		sort(nodes.begin(), nodes.end(), SortDataLinkOutAsc);
	}
	else if(m_orderBy == 2 && m_sortBy == 1)
	{
		sort(nodes.begin(), nodes.end(), SortDataLinkOutDesc);
	}
	else if(m_orderBy == 3 && m_sortBy == 0)
	{
		sort(nodes.begin(), nodes.end(), SortDataLinkInAsc);
	}
	else if(m_orderBy == 3 && m_sortBy == 1)
	{
		sort(nodes.begin(), nodes.end(), SortDataLinkInDesc);
	}
	else if(m_orderBy == 4 && m_sortBy == 0)
	{
		sort(nodes.begin(), nodes.end(), SortDataLinkBiAsc);
	}
	else if(m_orderBy == 4 && m_sortBy == 1)
	{
		sort(nodes.begin(), nodes.end(), SortDataLinkBiDesc);
	}
}

void Tree::SetSortOrder(int sortBy, int orderBy)
{
	m_sortBy = sortBy;
	m_orderBy = orderBy;
}

void Tree::SetTooltip(DataNode& dataNode, NodeRep& node)
{
	int inLink, outLink, biLink;
	GetRelations(dataNode, inLink, outLink, biLink);
	//QString toolTip;
	ostringstream toolTip;
	toolTip << "Name: " << dataNode.GetName().toStdString() << "\nOut: " << outLink << "\nIn: " << inLink << "\nBi: " << biLink;
	node.setToolTip(QString(toolTip.str().c_str()));
}

int Tree::GetLinkDirection(TreeNodeImpl *node)
{
	TreeNodeImpl *parent = static_cast<TreeNodeImpl *>(node->GetParentNode());
	return m_graph.GetLinkDirection(parent->GetDataNode(), node->GetDataNode());
}

/****************************************************************************\
* PROTECTED CLASS FUNCTIONS                                                  *
\****************************************************************************/

/****************************************************************************\
* PRIVATE CLASS FUNCTIONS                                                    *
\****************************************************************************/
void Tree::AddChild(TreeNode *node, DataNode& childData)
{
	TreeNode *child = m_nodeCatalog.Find(childData.GetId());
	if(child == NULL) 
	{
		child = new TreeNodeImpl(childData, *this, node);
		m_level[child] = m_level[node] + 1;
		m_maxDepth = max<int>(m_maxDepth, m_level[child]);
		node->AddChild(child);

		m_nodeCatalog.Insert(child->GetId(), child);
		m_app.GetScene()->addItem(&child->GetNodeRep());
		m_app.GetScene()->addItem(&child->GetLinkToParent()->GetTreeLinkRep());
	}
	else if(!IsChildParentOfNode(node, child))
	{
		node->AddChild(child);
		m_level[child] = m_level[node] + 1;
		m_maxDepth = max<int>(m_maxDepth, m_level[child]);
		child->SetParentNode(node);
	}
}

bool Tree::IsChildParentOfNode(TreeNode *node, TreeNode *child)
{
	if(node == child)
	{
		return true;
	}
	TreeNode *parent = node->GetParentNode();
	while(parent != NULL)
	{
		if(child == parent)
		{
			return true;
		}
		parent = parent->GetParentNode();
	}
	return false;
}

void Tree::RemoveChildren(TreeNode *node)
{
	list<TreeNode *> children;
	node->GetChildren(children);

	for(list<TreeNode *>::iterator it = children.begin(); it != children.end(); it++)
	{
		RemoveChildren(*it);
		node->RemoveChild(*it);
		RemoveNode(*it);
	}
//	while(!children.empty()) {
//		RemoveNode(children.front());
//		children.pop_front();
//	}
}

void Tree::RemoveNode(TreeNode *node)
{
	m_nodeCatalog.Remove(node->GetId());
	m_level.erase(node);
	delete (node);
}

void Tree::SetLevel(TreeNode *node, int level)
{
	m_level[node] = level;
	m_maxDepth = max<int>(m_maxDepth, level);
	
	list<TreeNode *> children;
	node->GetChildren(children);
	for(list<TreeNode *>::iterator it=children.begin(); it!=children.end(); it++)
	{
		SetLevel(*it, level + 1);
	}
}

void Tree::SetAnimationDuration(int ms)
{
	m_animationDuration = ms;
}

void Tree::SetMovedTo(TreeNodeImpl *oldParent, TreeNodeImpl *node)
{
	//remove movedToNodes from new parent
	TreeNodeImpl *newParent = static_cast<TreeNodeImpl *>(node->GetParentNode());
	if(newParent != NULL)
	{
		TreeNodeMovedTo *newMovedToParent = newParent->GetNodeMovedTo();
		if(newMovedToParent != NULL)
		{
			newMovedToParent->RemoveFromMovedToNodes(node);
			m_movedToParents[node].remove(newMovedToParent);
			if(m_movedToParents[node].size() == 0)
			{
				m_movedToParents.erase(node);
			}
			if(newMovedToParent->GetSizeMovedToNodes() == 0)
			{
				newParent->RemoveChild(newMovedToParent);
				RemoveNode(newMovedToParent);
			}
		}
	}

	//add movedToNodes to oldParent
	if(oldParent != node->GetParentNode())
	{
		// id could be 0
		int movedToId = (1 + oldParent->GetId()) * -1;
		TreeNodeMovedTo *movedToNode = static_cast<TreeNodeMovedTo *>(m_nodeCatalog.Find(movedToId));
		if(movedToNode == NULL)
		{
			movedToNode = new TreeNodeMovedTo(*this, oldParent, movedToId);
			movedToNode->AddMovedToNodes(node);
			m_movedToParents[node].push_back(movedToNode);
			oldParent->AddChild(movedToNode);
			m_nodeCatalog.Insert(movedToId, movedToNode);
			m_app.GetScene()->addItem(&movedToNode->GetNodeRep());
			m_app.GetScene()->addItem(&movedToNode->GetLinkToParent()->GetTreeLinkRep());
			movedToNode->SetPosition(oldParent->GetPosition()[0], oldParent->GetPosition()[1]);
			
			SetLevel(movedToNode, m_level[oldParent] + 1);
			movedToNode->SetAnimationGroup(ANIMATE_TREE_FIXED);
		}
		else 
		{
			movedToNode->AddMovedToNodes(node);
			m_movedToParents[node].push_back(movedToNode);
		}
	}
}

void Tree::RemoveFromMovedToNode(TreeNode *node)
{
	list<TreeNode *> children;
	node->GetChildren(children);
	for(list<TreeNode *>::iterator it = children.begin(); it != children.end(); it++)
	{
		if((*it)->GetType() == TYPE_NODE)
		{
			TreeNodeImpl *nodeImpl = static_cast<TreeNodeImpl *>(*it);
			RemoveFromMovedToNode(*it);
			if(m_movedToParents.find(nodeImpl) != m_movedToParents.end()) 
			{
				for(list<TreeNodeMovedTo *>::iterator itMoveTo = m_movedToParents[nodeImpl].begin(); itMoveTo != m_movedToParents[nodeImpl].end(); itMoveTo++)
				{
					(*itMoveTo)->RemoveFromMovedToNodes(nodeImpl);	
				}
				m_movedToParents.erase(nodeImpl);
			}
		}
	}
}

void Tree::RemoveMovedTo(TreeNodeImpl *node)
{
	TreeNodeMovedTo *movedTo = node->GetNodeMovedTo();
	if(movedTo != NULL && movedTo->GetSizeMovedToNodes() == 0)
	{
		TreeNode *parent = movedTo->GetParentNode();
		parent->RemoveChild(movedTo);
		RemoveNode(movedTo);
	}
	list<TreeNode *> children;
	node->GetChildren(children);
	for(list<TreeNode *>::iterator it=children.begin(); it != children.end(); it++)
	{
		if((*it)->GetType() == TYPE_NODE)
		{
			RemoveMovedTo(static_cast<TreeNodeImpl *>(*it));
		}
	}
}

void Tree::GetRelations(DataNode& node, int &inNodesNum, int &outNodesNum, int &biNodesNum)
{

	//get all outgoing nodes
	vector<DataNode*> tmpOutNodes;
	m_graph.GetNodesOut(node, tmpOutNodes);

	//get all incoming nodes
	vector<DataNode*> tmpInNodes;
	m_graph.GetNodesIn(node, tmpInNodes);

	//set bidirectional nodes
	std::vector<DataNode *> biNodes;
	set_intersection(tmpInNodes.begin(), tmpInNodes.end(), tmpOutNodes.begin(), tmpOutNodes.end(), inserter(biNodes, biNodes.begin()));
	
	//set outgoing nodes (outgoing without bidirectional)
	std::vector<DataNode *> outNodes;
	set_difference(tmpOutNodes.begin(), tmpOutNodes.end(), biNodes.begin(), biNodes.end(), inserter(outNodes, outNodes.begin()));
	
	//set incoming nodes (incoming without bidirectional)
	std::vector<DataNode *> inNodes;
	set_difference(tmpInNodes.begin(), tmpInNodes.end(), biNodes.begin(), biNodes.end(), inserter(inNodes, inNodes.begin()));
	
	inNodesNum = inNodes.size();
	outNodesNum = outNodes.size();
	biNodesNum = biNodes.size();
}


