/****************************************************************************\
*                                                                            *
* 02.06.2008                                                                 *
* Vector2 - Header File                                                      *
*                                                                            *
* (c) Robert Hausmair										                 *
*                                                                            *
\****************************************************************************/
#ifndef __VECTOR2_H__
#define __VECTOR2_H__

/********************************\
* REQUIRED HEADERS               *
\********************************/

#include <Definitions.h>
#include <cmath>

/****************************************************************************\
* TEMPLATE STRUCT                                                            *
\****************************************************************************/
template<typename T>
struct Vector2
{
	/************************************************************************/
	public:// PUBLIC CLASS FUNCTIONS                                        */
	/************************************************************************/

		//STATIC-------------------------------------------------------------

		static Vector2<T> Interpolate(Vector2<T> from, Vector2<T> to, T val)
		{
			T x = ((1.0 - val) * from.m_X) + (val * to.m_X);
			T y = ((1.0 - val) * from.m_Y) + (val * to.m_Y);
			return Vector2<T>(x,y);
		}
		//CONSTRUCTOR--------------------------------------------------------

	    Vector2() {}     //No initialization for speed purposes
		Vector2(const T x, const T y) 
			: m_X(x), m_Y(y) { }
		Vector2(const Vector2<T> &other)
			: m_X(other.m_X), m_Y(other.m_Y) { }

		//OPERATORS----------------------------------------------------------

		//assign
		const Vector2<T> &operator=(const Vector2<T> &other) 
		{
			m_X = other.m_X; m_Y = other.m_Y; 
			return *this;
		}

		//self assign
		const Vector2<T> &operator+=(const Vector2<T> &other) 
		{
			m_X += other.m_X; m_Y += other.m_Y; 
			return *this;
		}
		const Vector2<T> &operator-=(const Vector2<T> &other) 
		{
			m_X -= other.m_X; m_Y -= other.m_Y; 
			return *this;
		}
		const Vector2<T> &operator*=(const T factor) 
		{
			m_X *= factor; m_Y *= factor; 
			return *this;
		}
		const Vector2<T> &operator/=(const T quot) 
		{
			m_X /= quot; m_Y /= quot; 
			return *this;
		}

		//binary
		const Vector2<T> operator+(const Vector2<T> &other) const 
		{
			return Vector2<T>(m_X + other.m_X, m_Y + other.m_Y);
		}
		const Vector2<T> operator-(const Vector2<T> &other) const 
		{
			return Vector2<T>(m_X - other.m_X, m_Y - other.m_Y);
		}
		const Vector2<T> operator*(const T factor) const 
		{
			return Vector2<T>(m_X * factor, m_Y * factor);
		}
		const Vector2<T> operator/(const T quot) const 
		{
			return Vector2<T>(m_X / quot, m_Y / quot);
		}

		//negate
		const Vector2<T> operator-() const 
		{
			return Vector2<T>(-m_X, -m_Y);
		}

		//compare	
		bool operator==(const Vector2<T> &other) 
		{
			return m_X == other.m_X && m_Y == other.m_Y;
		}
		bool operator!=(const Vector2<T> &other) 
		{
			return m_X != other.m_X || m_Y != other.m_Y;
		}

		//index
		T &operator[](const uchar i) 
		{
			assert(i < 2 && "Error: Invalid index for Vector2::operator[]!"); 
			return *(&m_X + i);
		}
		const T &operator[](const uchar i) const 
		{
			assert(i < 2 && "Error: Invalid index for Vector2::operator[]!"); 
			return *(&m_X + i);
		}

		//explicit set
		void Set(const T x, const T y) 
		{
			m_X = x; m_Y = y;
		}

		//MEMBER MATH-------------------------------------------------------------

		const double Length() const 
		{
			return sqrt(m_X * m_X + m_Y * m_Y);
		}
		const T LengthSqr() const 
		{
			return m_X * m_X + m_Y * m_Y;
		}
		const T Dot(const Vector2<T> &other) const 
		{
			return m_X * other.m_X + m_Y * other.m_Y
		}

		void Normalize() 
		{
			assert(LengthSqr() && "Error: Vector2::Normalize called for a zero vector!"); 
			float fL = 1 / Length(); 
			m_X *= fL; m_Y *= fL;
		}
		Vector2<T> Unit() const 
		{
			float fL = 1.0 / Length(); 
			return Vector2<T>(m_X*fL,m_Y*fL);
		}
	
		//LOGGING------------------------------------------------------------------

		//void Log(ushort wTargetLog) const {LOG(wTargetLog, "(%.2f, %.2f)", m_X, m_Y);}

	/***************************************************************************/
	public: //PUBLIC CLASS MEMBERS                                             */
	/***************************************************************************/
    
		T m_X, m_Y;
};

/********************************\
* DEFINITIONS                    *
\********************************/

typedef Vector2<float> Vector2f;
typedef Vector2<double> Vector2d;
typedef Vector2<long> Vector2l;
typedef Vector2<int> Vector2i;
typedef Vector2<short> Vector2s;

#endif