/****************************************************************************\
*                                                                            *
* 02.06.2008                                                                 *
* Vector3 - Header File                                                      *
*                                                                            *
* (c) Robert Hausmair										                 *
*                                                                            *
\****************************************************************************/
#ifndef __VECTOR3_H__
#define __VECTOR3_H__


/********************************\
* REQUIRED HEADERS               *
\********************************/

#include <Definitions.h>
#include <cmath>

/****************************************************************************\
* TEMPLATE STRUCT                                                            *
\****************************************************************************/
template<typename T>
struct Vector3
{
	/***************************************************************************/
	public: //PUBLIC CLASS FUNCTIONS                                           */
	/***************************************************************************/
 	
		static const Vector3 ZERO;
		static const Vector3 UNIT_X;
		static const Vector3 UNIT_Y;
		static const Vector3 UNIT_Z;

		//STATIC-------------------------------------------------------------

		static Vector3<T> Interpolate(Vector3<T> from, Vector3<T> to, T val)
		{
			float oneMinusVal = 1.0 - val;
			T x = (oneMinusVal * from.m_X) + (val * to.m_X);
			T y = (oneMinusVal * from.m_Y) + (val * to.m_Y);
			T z = (oneMinusVal * from.m_Z) + (val * to.m_Z);
			return Vector3<T>(x,y,z);
		}

		//CONSTRUCTOR--------------------------------------------------------

		Vector3(){}//No initialization for speed purposes    
		Vector3(const T x, const T y, const T z) : m_X(x), m_Y(y), m_Z(z) {}
		Vector3(const Vector3<T> &other) : m_X(other.m_X), m_Y(other.m_Y), m_Z(other.m_Z) {}

		//OPERATORS----------------------------------------------------------

		//assign
		const Vector3<T> &operator =(const Vector3<T> &other)
		{
			m_X = other.m_X; m_Y = other.m_Y; m_Z = other.m_Z; 
			return *this;
		}

		//self assign
		const Vector3<T> &operator +=(const Vector3<T> &other) 
		{
			m_X += other.m_X; m_Y += other.m_Y; m_Z += other.m_Z; 
			return *this;
		}
		const Vector3<T> &operator -=(const Vector3<T> &other) 
		{
			m_X -= other.m_X; m_Y -= other.m_Y; m_Z -= other.m_Z; 
			return *this;
		}
		const Vector3<T> &operator *=(const T factor) 
		{
			m_X *= factor; m_Y *= factor; m_Z *= factor; 
			return *this;
		}
		const Vector3<T> &operator /=(const T quot) {
			m_X /= quot; m_Y /= quot; m_Z /= quot; 
			return *this;
		}

		//compare
		bool operator ==(const Vector3<T> &other)
		{
			return m_X == other.m_X && m_Y == other.m_Y && m_Z == other.m_Z;
		}
		bool operator !=(const Vector3<T> &other) 
		{
			return m_X != other.m_X || m_Y != other.m_Y || m_Z != other.m_Z;
		}

		//binary
		const Vector3<T> operator +(const Vector3<T> &other) const 
		{
			return Vector3<T>(m_X + other.m_X, m_Y + other.m_Y, m_Z + other.m_Z);
		}
		const Vector3<T> operator -(const Vector3<T> &other) const 
		{
			return Vector3<T>(m_X - other.m_X, m_Y - other.m_Y, m_Z - other.m_Z);
		}
		const Vector3<T> operator *(const T factor) const 
		{
			return Vector3<T>(m_X * factor, m_Y * factor, m_Z * factor);
		}
		const Vector3<T> operator /(const T quot) const 
		{
			return Vector3<T>(m_X / quot, m_Y / quot, m_Z / quot);
		}
	   
		//negate
		Vector3<T> operator -() const 
		{
			return Vector3<T>(-m_X, -m_Y, -m_Z);
		}
		
		//index
		T & operator[](const uchar i) 
		{
			assert(i < 3 && "Error: Invalid index for Vector3::operator[]!");
			return *(&m_X + i);
		}
		const T &operator[](const uchar i) const 
		{
			assert(i < 3 && "Error: Invalid index for Vector3::operator[]!");
			return *(&m_X + i);
		}

		//cross product
		const Vector3<T> operator *(const Vector3<T> &other) const 
		{
			return Vector3<T>(m_Y * other.m_Z - m_Z * other.m_Y, m_Z * other.m_X - m_X * other.m_Z, m_X * other.m_Y - m_Y * other.m_X);
		}

		//explicit set
		void Set(const T x, const T y, const T z) 
		{
			m_X = x; m_Y = y; m_Z = z;
		}

		//MEMBER MATH-------------------------------------------------------------

		const float Length() const 
		{
			return sqrt(m_X * m_X + m_Y * m_Y + m_Z * m_Z);
		}
		const float LengthSqr() const 
		{
			return m_X * m_X + m_Y * m_Y + m_Z * m_Z;
		}
		bool IsZero() const 
		{
			return((m_X == 0.0f) && (m_Y == 0.0f) && m_Z == 0.0f);
		}
		const T Dot(const Vector3<T> &other) const 
		{
			return m_X * other.m_X + m_Y * other.m_Y + m_Z * other.m_Z;
		}
		void Normalize() 
		{
			assert(LengthSqr() && "Error: Vector3::Normalize called for a zero vector!");
			float fL = 1 / Length(); 
			m_X *= fL; m_Y *= fL; m_Z *= fL;
		}
		Vector3<T> Unit() const 
		{
			assert(LengthSqr() && "Error: Vector3::Normalize called for a zero vector!");
			float fL = 1.0 / Length(); 
			return Vector3<T>(m_X*fL,m_Y*fL,m_Z*fL);
		}

	 /***************************************************************************/
	 public: //PUBLIC CLASS MEMBERS                                             */
	 /***************************************************************************/
	   
			T m_X, m_Y, m_Z;
};

/***********************************\
* CONSTANTS               	        *
\***********************************/

template<typename T>
const Vector3<T> Vector3<T>::ZERO(0.0,0.0,0.0);

template<typename T>
const Vector3<T> Vector3<T>::UNIT_X(1.0,0.0,0.0);

template<typename T>
const Vector3<T> Vector3<T>::UNIT_Y(0.0,1.0,0.0);

template<typename T>
const Vector3<T> Vector3<T>::UNIT_Z(0.0,0.0,1.0);

/***********************************\
* DEFINITIONS	                    *
\***********************************/

typedef Vector3<float> Vector3f;
typedef Vector3<double> Vector3d;
typedef Vector3<short> Vector3s;
typedef Vector3<long> Vector3l;
typedef Vector3<int> Vector3i;

#endif