/****************************************************************************\
*                                                                            *
* 02.06.2008                                                                 *
* Vector4 - Header File                                                      *
*                                                                            *
* (c) Robert Hausmair										                 *
*                                                                            *
\****************************************************************************/
#ifndef __VECTOR4_H__
#define __VECTOR4_H__

/********************************\
* REQUIRED HEADERS               *
\********************************/

#include <Definitions.h>
#include <Math/Vector2.h>
#include <Math/Vector3.h>
#include <cmath>

/****************************************************************************\
* TEMPLATE STRUCT                                                            *
\****************************************************************************/
template<typename T>
struct Vector4
{
	/************************************************************************/
	public:// PUBLIC CLASS FUNCTIONS                                        */
	/************************************************************************/
    
		static const Vector4 ZERO;
		static const Vector4 UNIT_X;
		static const Vector4 UNIT_Y;
		static const Vector4 UNIT_Z;

		//CONSTRUCTOR--------------------------------------------------------

        Vector4() {}//No initialization for speed purposes
        Vector4(const T x, const T y, const T z, const T w) : m_X(x), m_Y(y), m_Z(z), m_W(w) {}
		Vector4(const Vector4<T> &other) : m_X(other.m_X), m_Y(other.m_Y), m_Z(other.m_Z), m_W(other.m_W) {}

		//OPERATORS----------------------------------------------------------

		//assign
		const Vector4<T> &operator =(const Vector4<T> &other) 
		{
			m_X = other.m_X; m_Y = other.m_Y; m_Z = other.m_Z; m_W = other.m_W; 
			return *this;
		}
        const Vector4<T> &operator =(const Vector3<T> &other) 
		{
			m_X = other.m_X; m_Y = other.m_Y; m_Z = other.m_Z; m_W = 1.0f; 
			return *this;
		}
		const Vector4<T> &operator =(const Vector2<T> &other) 
		{
			m_X = other.m_X; m_Y = other.m_Y; m_Z = 0.0f; m_W = 1.0f; 
			return *this;
		}

		//self assign
		const Vector4<T> &operator +=(const Vector4<T> &other) 
		{
			m_X += other.m_X; m_Y += other.m_Y; m_Z += other.m_Z; m_W += other.m_W;
			return *this;
		}
		const Vector4<T> &operator -=(const Vector4<T> &other) 
		{
			m_X -= other.m_X; m_Y -= other.m_Y; m_Z -= other.m_Z; m_W -= other.m_W; 
			return *this;
		}
		const Vector4<T> &operator *=(const T factor) 
		{
			m_X *= factor; m_Y *= factor; m_Z *= factor; m_W *= factor; 
			return *this;
		}
		const Vector4<T> &operator /=(const T quot) 
		{
			m_X /= quot; m_Y /= quot; m_Z /= quot; m_W /= quot; 
			return *this;
		}

		//compare
		bool operator ==(const Vector4<T> &other) 
		{
			return m_X == other.m_X && m_Y == other.m_Y && m_Z == other.m_Z && m_W == other.m_W;
		}
		bool operator !=(const Vector4<T> &other) 
		{
			return m_X != other.m_X || m_Y != other.m_Y || m_Z != other.m_Z || m_W != other.m_W;
		}

		//binary
		const Vector4<T> operator +(const Vector4<T> &other) const 
		{
			return Vector4<T>(m_X + other.m_X, m_Y + other.m_Y, m_Z + other.m_Z, m_W + other.m_W);
		}
        const Vector4<T> operator -(const Vector4<T> &other) const 
		{
			return Vector4<T>(m_X - other.m_X, m_Y - other.m_Y, m_Z - other.m_Z, m_W - other.m_W);
		}
		const Vector4<T> operator *(const T factor) const 
		{
			return Vector4<T>(m_X * factor, m_Y * factor, m_Z * factor, m_W * factor);
		}
		const Vector4<T> operator /(const T quot) const 
		{
			return Vector4<T>(m_X / quot, m_Y / quot, m_Z / quot, m_W / quot);
		}
		
		//index
        T &operator[](const uchar i) 
		{
			assert(i < 4 && "Error: Invalid index for Vector4::operator[]!"); 
			return *(&m_X + i);
		}
        const T &operator[](const uchar i) const 
		{
			assert(i < 4 && "Error: Invalid index for Vector4::operator[]!"); 
			return *(&m_X + i);
		}

		//explicit set
		 void Set(const T x, const T y, const T z, const T w) 
		 {
			 m_X = x; m_Y = y; m_Z = z; m_W = w;
		 }

		//MEMBER MATH-------------------------------------------------------------

        const float Length() const 
		{
			return sqrt(m_X * m_X + m_Y * m_Y + m_Z * m_Z + m_W * m_W);
		}
        const float LengthSqr() const 
		{
			return m_X * m_X + m_Y * m_Y + m_Z * m_Z + m_W * m_W;
		}
		const T Dot(const Vector4<T> &other) const 
		{
			return m_X * other.m_X + m_Y * other.m_Y + m_Z * other.m_Z + m_W * other.m_W;
		}
		const Vector4<T> ElementMultiply(const Vector4<T> &other) const 
		{
			return Vector4<T>(m_X * other.m_X, m_Y * other.m_Y, m_Z * other.m_Z, m_W * other.m_W);
		}
        void Normalize() 
		{
			assert(LengthSqr() && "Error: Vector4::Normalize called for a zero vector!");
			float fL = 1 / Length(); m_X *= fL; m_Y *= fL; m_Z *= fL; m_W *= fL;
		}
		Vector4<T> Unit() const 
		{
			assert(LengthSqr() && "Error: Vector4::Normalize called for a zero vector!");
			float fL = 1.0 / Length(); 
			return Vector4<T>(m_X*fL, m_Y*fL, m_Z*fL, m_W*fL);
		}
      
	/************************************************************************/
	public:// PUBLIC CLASS MEMBERS                                          */
	/************************************************************************/
    
        T m_X, m_Y, m_Z, m_W;
};

/***********************************\
* CONSTANTS               	        *
\***********************************/

template<typename T>
const Vector4<T> Vector4<T>::ZERO(0.0,0.0,0.0,0.0);

template<typename T>
const Vector4<T> Vector4<T>::UNIT_X(1.0,0.0,0.0,0.0);

template<typename T>
const Vector4<T> Vector4<T>::UNIT_Y(0.0,1.0,0.0,0.0);

template<typename T>
const Vector4<T> Vector4<T>::UNIT_Z(0.0,0.0,1.0,0.0);

/********************************\
* DEFINITIONS                    *
\********************************/

typedef Vector4<float> Vector4f;
typedef Vector4<double> Vector4d;
typedef Vector4<short> Vector4s;
typedef Vector4<long> Vector4l;
typedef Vector4<int> Vector4i;

#endif