/******************************************************************************\
*                                                                              *
* 1/4/2009                                                                     *
* TreePlus - Source File                                                       *
*                                                                              *
* (c) Robert Hausmair, Christoph Winklhofer                                    *
*                                                                              *
\******************************************************************************/

//REQUIRED HEADERS 

#include <TreePlus.h>
#include <Scene.h>
#include <View.h>
#include <Xml/ParseGraph.h>
#include <Graph/Graph.h>
#include <Graph/Tree.h>
#include <Layout/TreeLayout.h>
#include <QFileDialog>
#include <QCloseEvent>
#include <Memory/mmgr.h>

//NAMESPACE

using namespace std;

//PRIVATE DEFINITIONS TYPEDEFS

//

/****************************************************************************\
* STATIC                                                                     *
\****************************************************************************/

//MEMBER

//METHODS

/****************************************************************************\
* CD                                                                         *
\****************************************************************************/

TreePlus::TreePlus()
{
	//init
	m_pTree = NULL;
	m_pGraph = NULL;

	//setup the user intrface
	m_ui.setupUi(this);

	//connect menu actions
	connect(menuBar(),SIGNAL(triggered(QAction *)),this,SLOT(MenuTriggered(QAction *)));
	
	//connect settings apply button
	connect(&m_settings,SIGNAL(Apply()),this,SLOT(ApplySettings()));
}

TreePlus::~TreePlus()
{
	SAFE_DELETE(m_pTree);
	SAFE_DELETE(m_pGraph);
}

/****************************************************************************\
* PUBLIC CLASS FUNCTIONS                                                     *
\****************************************************************************/

//CREATE
bool TreePlus::Create()
{
	//INIT
	NodeRep::Init();
	TreeNodeRep::Init();
	PreviewNodeRep::Init();

	//create the scene
	m_pScene = new Scene(this);
	m_pScene->setSceneRect(-50000,-50000,100000,100000);
			
	//create the view
	m_pView = new View(this);
	m_pView->setMinimumWidth(500);
	m_pView->setSizePolicy(QSizePolicy::Expanding,QSizePolicy::Preferred);
	m_pView->SetScaleRange(0.2f,1.5f);
	m_pView->setScene(m_pScene);
	
	//attach view to central widget
	this->setCentralWidget(m_pView);

	//ParseGraph parseGraph("../data/sample_foodweb.graphml");
	//m_pGraph = parseGraph.CreateGraph();
	//m_pTree = new Tree(*m_pGraph,*this);
	
	//apply settings
	ApplySettings();
	
	return true;
}

/****************************************************************************\
* PROTECTED CLASS FUNCTIONS                                                  *
\****************************************************************************/

void TreePlus::closeEvent(QCloseEvent *event)
{
	m_settings.close();
	event->accept();
}

/****************************************************************************\
* PRIVATE CLASS FUNCTIONS                                                    *
\****************************************************************************/

//MENU TRIGGERED
void TreePlus::MenuTriggered(QAction *action)
{
	if(action->text().compare("Open") == 0)
	{
		//create and show file dialog
		QString fileName = QFileDialog::getOpenFileName(this, "Open file", "./data", "GraphML File (*.graphml)");

		if(!fileName.isNull())
		{
			//delete old
			if(m_pGraph)
			{
				delete m_pTree;
				delete m_pGraph;
			}

			//parse and create
			ParseGraph parseGraph(fileName);
			m_pGraph = parseGraph.CreateGraph();
			m_pTree = new Tree(*m_pGraph,*this);

			ApplySettings();
			m_pView->SetScaleRange(0.2f,1.5f);
			m_pTree->ExploreNode(static_cast<TreeNodeImpl *>(m_pTree->GetRoot()));
			m_pTree->SetLayout();
		}
		return;
	}

	if(action->text().compare("Settings") == 0)
	{
		if(!m_settings.isVisible())
		{
			//center settings window
			QRect pGeom = this->geometry();
			QSize sSize = m_settings.size();
			QPoint sPoint = QPoint( (pGeom.width()-sSize.width())/2.0 , (pGeom.height()-sSize.height())/2.0 );
			m_settings.setGeometry(pGeom.x()+sPoint.x(), pGeom.y()+sPoint.y(),sSize.width(), sSize.height());

			m_settings.setModal(true);
			m_settings.setWindowModality(Qt::WindowModal);
			m_settings.setWindowFlags(Qt::WindowStaysOnTopHint);

			//show window modal
			if(m_settings.exec())
			{
				ApplySettings();
			}
		}
		return;
	}

	if(action->text().compare("Exit") == 0)
	{
		Exit();
		return;
	}
}

//EXIT
void TreePlus::Exit()
{
	m_settings.close();
	qApp->quit();
}

//APPLY SETTINGS
void TreePlus::ApplySettings()
{
	//get
	m_settings.GetData(m_settingsData);
	
	//apply nodeSize
	NodeRep::SetNodeSize(m_settingsData.m_nodeWidth,m_settingsData.m_nodeHeight);
	
	//adjust Preview
	if(m_pView->GetPreview())
	{
		m_pView->GetPreview()->Adjust(m_settingsData.m_nodeWidth,m_settingsData.m_nodeHeight);
	}
		
	//apply nodeColors
	NodeRep::SetNodeColor
	(
		m_settingsData.m_colNodeIngoing,
		m_settingsData.m_colNodeOutgoing,
		m_settingsData.m_colNodeBidirectional,
		m_settingsData.m_colNodeSelected
	);

	//apply nodeLabel settings
	NodeRep::SetNodeLabelFont(m_settingsData.m_fontSize,m_settingsData.m_colNodeLabels);

	//apply treeNode settings
	TreeNodeRep::SetComponents(m_settingsData.m_drawArrows,m_settingsData.m_drawInfofields);
	
	//apply previewBg color
	m_pView->SetPreviewBgColor(m_settingsData.m_colPreviewBg);

	//apply sceneBg color
	m_pView->SetSceneBgColor(m_settingsData.m_colSceneBg);

	if(m_pTree != NULL)
	{
		//apply animationDuration
		m_pTree->SetAnimationDuration(m_settingsData.m_animationDuration);

		//apply sort order
		m_pTree->SetSortOrder(m_settingsData.m_sortBy, m_settingsData.m_orderBy);

		//apply camera animation
		m_pTree->SetAnimateView(m_settingsData.m_animateCamera);

		//redraw
		m_pTree->RedrawNodes();
	}

	//apply treeLayout
	TreeLayout::SetLevelSeparation(m_settingsData.m_levelSpacing + m_settingsData.m_nodeWidth);
	TreeLayout::SetSiblingSeparation(m_settingsData.m_siblingSpacing);
	TreeLayout::SetSubtreeSeparation(m_settingsData.m_subtreeSpacing);

	//check layout change
	if(m_settingsData.m_bLayoutChange)
	{
		//update layout
		if(m_pTree != NULL)
		{
			m_pTree->UpdateLayout();
		}
	}

	//repaint
	this->activateWindow();
}
