/******************************************************************************\
*                                                                              *
* 1/4/2009                                                                     *
* View - Source File                                                           *
*                                                                              *
* (c) Robert Hausmair, Christoph Winklhofer                                    *
*                                                                              *
\******************************************************************************/

//REQUIRED HEADERS 

#include <View.h>
#include <QGLWidget>
#include <QResizeEvent>
#include <Memory/mmgr.h>

//NAMESPACE

using namespace std;

//PRIVATE DEFINITIONS TYPEDEFS

/****************************************************************************\
* STATIC                                                                     *
\****************************************************************************/

//MEMBER

//METHODS

/****************************************************************************\
* CD                                                                         *
\****************************************************************************/

View::View(QWidget *parent) : 
	m_pPreview(NULL),
	QGraphicsView(parent) 
{
	//init scale
	m_currentScale = 1.0f;
		
	//init ScaleRange
	SetScaleRange(0.3f,2.0f);

	//enable OpenGL rendering
	setViewport(new QGLWidget(QGLFormat(QGL::SampleBuffers)));

	//full update when using OpenGL rendering
	setViewportUpdateMode(QGraphicsView::FullViewportUpdate);

	//set appropriate anchor
	setResizeAnchor(QGraphicsView::AnchorUnderMouse);
	setTransformationAnchor(QGraphicsView::AnchorViewCenter);

	//set dragMode
	setDragMode(ScrollHandDrag);

	//disable scrollbars
	setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
	setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);

	//renderHint
	setRenderHints(QPainter::Antialiasing | QPainter::SmoothPixmapTransform);

	//init preview
	SetPreviewBgColor(QColor(220,220,200));
	m_visibleFrame = NULL;
}

View::~View()
{
	
}

/****************************************************************************\
* PUBLIC CLASS FUNCTIONS                                                     *
\****************************************************************************/

//SET PREVIEW BG COLOR
void View::SetPreviewBgColor( QColor &col )
{
	//set
	m_previewBgColor = col;

	//check
	if(m_pPreview)
	{
		QLinearGradient gradient(QPointF(0,0),QPointF(m_pPreview->boundingRect().width(),0));
		gradient.setColorAt(0,QColor(100,100,100,255));
		gradient.setColorAt(0.006,QColor(255,255,255,255));
		gradient.setColorAt(0.016,m_previewBgColor);
		gradient.setColorAt(1.0,m_previewBgColor);
		m_pPreview->setBrush(QBrush(gradient));
	}
}

//SET SCENE BG COLOR
void View::SetSceneBgColor( QColor &col )
{
	scene()->setBackgroundBrush(QBrush(col));
}

//SET SCALE RANGE
void View::SetScaleRange(float scaleMin, float scaleMax)
{
	m_minScale = scaleMin;
	m_maxScale = scaleMax;
}

//LOCK INPUT
void View::LockInput(bool tf)
{
	setInteractive(!tf);
}

//LOOK AT
void View::LookAt( QRectF &frame )
{
	fitInView(frame,Qt::KeepAspectRatio);

	//update currentScale
	m_currentScale = transform().m11();

	DoScale(m_currentScale);
}

//COMPUTE To RECT
QRectF View::ComputeToRect( QRectF &frame, QPointF &nodePos )
{
	//display
	//m_visibleFrame->setRect(frame);

	//fit
	QPointF w0(this->mapToScene(QPoint(0,0)));
	QPointF w1(this->mapToScene(QPoint(width(),0)));
	QPointF wScene = w1-w0;

	//ADJUST
	QRectF fitFrame = QRectF(frame);
	fitFrame.setWidth(wScene.x()*m_currentScale);
	fitFrame.adjust(-200*m_currentScale,-200*m_currentScale,200*m_currentScale,200*m_currentScale);

	//STORE
	QRectF backupRect  = mapToScene(this->viewport()->rect()).boundingRect();

		//APPLY
		fitInView(fitFrame,Qt::KeepAspectRatio);

		QPointF centerPoint;
		if((frame.right() - nodePos.x()) > NodeRep::GetNodeSize().width())
		{
			//centerPoint = QPointF(nodePos.x() + 0.5*NodeRep::GetNodeSize().width()/m_currentScale , nodePos.y() - 0.5*NodeRep::GetNodeSize().height()/m_currentScale );
		}
		else
		{
			//centerPoint = QPointF( frame.x() + frame.right() - 0.5*NodeRep::GetNodeSize().width()/m_currentScale, frame.y() + ((frame.bottom()-frame.top())/2.0f) );
		}

		centerPoint = centerPoint = QPointF( frame.x() + frame.right() - 0.5*NodeRep::GetNodeSize().width()/m_currentScale, frame.y() + ((frame.bottom()-frame.top())/2.0f) );
		centerOn(centerPoint);
	
		//RECORD
		QRectF result = mapToScene(this->viewport()->rect()).boundingRect();

	//RESTORE
	fitInView(backupRect,Qt::KeepAspectRatio);

	//return
	return result;
}

//GET VIEW RECT
QRectF View::GetViewRect()
{
	return this->mapToScene(this->viewport()->rect()).boundingRect();
}

/****************************************************************************\
* PROTECTED CLASS FUNCTIONS                                                  *
\****************************************************************************/

//PAINT EVENT
void View::paintEvent( QPaintEvent *event )
{
	QRectF previewRect = m_pPreview->boundingRect();
	QPointF previewRepScenePos = this->mapToScene(QPoint(width() - previewRect.width(), 0));
	
	m_pPreview->setPos(previewRepScenePos);
	m_pPreview->setMatrix(QMatrix().scale(1.0f/m_currentScale,1.0f/m_currentScale));

	QGraphicsView::paintEvent(event);
}

// SHOW EVENT
void View::showEvent(QShowEvent *event)
{
	if(!m_pPreview)
	{
		m_pPreview = new Preview(*this);
		m_pPreview->setZValue(500000);
		SetPreviewBgColor(m_previewBgColor);
		scene()->addItem(m_pPreview);
		
		//m_visibleFrame = new QGraphicsRectItem();
		//scene()->addItem(m_visibleFrame);
	}

	QGraphicsView::showEvent(event);
}

//RESIZE EVENT
void View::resizeEvent( QResizeEvent *event )
{
	if(m_pPreview)
	{
		m_pPreview->UpdateGeometry(*this);
	}
	QGraphicsView::resizeEvent(event);
}

//MOUSE PRESS EVENT
static QPoint pressStart;
static float scaleStart = 0;
static bool buttonZoom = false;
void View::mousePressEvent(QMouseEvent *event)
{
	//remember
	pressStart = QPoint(event->x(),event->y());
	scaleStart = m_currentScale;

	//check if pressed in viewport
	QRectF previewRect = m_pPreview->boundingRect();
	QRectF previewViewRect = QRectF(width() - previewRect.width(),0,previewRect.width(),height());
	if(!previewViewRect.contains(event->posF()))
	{
		//set
		buttonZoom = true;
	}

	//forward
	QGraphicsView::mousePressEvent(event);
}

//MOUSE RELEASE EVENT
void View::mouseReleaseEvent(QMouseEvent *event)
{
	//reset
	buttonZoom = false;

	//forward
	QGraphicsView::mouseReleaseEvent(event);
}

//MOUSE MOVE EVENT
void View::mouseMoveEvent(QMouseEvent *event)
{
	//check if not in previewWindow
	QRectF previewRect = m_pPreview->boundingRect();
	QRectF previewViewRect = QRectF(width() - previewRect.width(),0,previewRect.width(),height());
	bool insidePreview = previewViewRect.contains(event->posF());

	if(event->buttons() == Qt::RightButton)
	{
		//zoom and dont forward
		if((!insidePreview && buttonZoom || insidePreview && buttonZoom))
		{
			//calculate
			QPoint pressEnd(event->x(),event->y());
			float dy = pressEnd.y() - pressStart.y() > 0 ? 1.0f : -1.0f;

			//newscale
			m_currentScale = scaleStart + ((pressEnd - pressStart).manhattanLength()*dy*0.0006); 

			//limit
			if(m_currentScale > m_maxScale)
			{
				m_currentScale = m_maxScale;
			}
			if(m_currentScale < m_minScale)
			{
				m_currentScale = m_minScale;
			}

			//scale
			DoScale(m_currentScale);
			event->accept();
			return;
		}
	}

	//forward to scene
	QGraphicsView::mouseMoveEvent(event);
}

//MOUSE DOUBLE CLICK
void View::mouseDoubleClickEvent( QMouseEvent *event )
{
	//forward
	QGraphicsView::mouseDoubleClickEvent(event);
}

//WHEEL EVENT
void View::wheelEvent( QWheelEvent *event )
{
	event->accept();
	return;
}

//VIEWPORT EVENT
bool View::viewportEvent( QEvent *event )
{
	return QGraphicsView::viewportEvent(event);
}

/****************************************************************************\
* PRIVATE CLASS FUNCTIONS                                                    *
\****************************************************************************/

//DO SCALE
void View::DoScale( float newScale )
{
	PreviewNodeRep::SetGlobalScale(newScale);
	setTransform(QTransform().scale(newScale,newScale));
}


